/*--------------------------------------------------------------------
|                                                                     |
|    Convoy                                                           |
|                                                                     |
 -------------------------------------------------------------------- 
|
| Author : GeekToo
| Version: 1
|
|
 --------------------------------------------------------------------*/

/*--------------------------------------------------------------------
|                                                                     |
|    Tile                                                             |
|                                                                     |
 --------------------------------------------------------------------*/
class Tile
{
	constructor() {
		
	}
	function GetAdjacentTiles(tile);
}	

function Tile::GetAdjacentTiles(tile)
{
	local adjTiles = AITileList();
	
	adjTiles.AddTile(tile - AIMap.GetTileIndex(1,0));
	adjTiles.AddTile(tile - AIMap.GetTileIndex(0,1));
	adjTiles.AddTile(tile - AIMap.GetTileIndex(-1,0));
	adjTiles.AddTile(tile - AIMap.GetTileIndex(0,-1));
	
	return adjTiles;
}

/*--------------------------------------------------------------------
|                                                                     |
|    Node                                                             |
|                                                                     |
 --------------------------------------------------------------------*/
class Node
{
	tile_index = 0;
	predec     = 0;
	f = 0;
	g = 0;
	h = 0;

	constructor() {
		tile_index = 0;
		predec     = 0;
		f = 0;
		g = 0;
		h = 0;
	}
}

/*--------------------------------------------------------------------
|                                                                     |
|    NodeList                                                         |
|                                                                     |
 --------------------------------------------------------------------*/
class NodeList
{
	list = null;
	constructor() {
		list = {};
	}
	function Append(node);
	function Exists(a);
}

function NodeList::Append(node)
{
	list.rawset(node.tile_index,node);
}


function NodeList::Exists(a)
{
	local node = -1;
	if (list.rawin(a)){
		node = list.rawget(a);
	}
	return node;
}



/*--------------------------------------------------------------------
|                                                                     |
|   BinHeap                                                           |
|                                                                     |
 --------------------------------------------------------------------*/
class BinHeap
{
	list = null;

	constructor() {
		list = [];
	}
	function Insert(node);
	function PopMin();
	function BubbleDown();
	function Exists(a);
	function Dump();
}

/*------------------------------------------------------------------*/
function BinHeap::Insert(node)
{
	node.f = node.g + node.h;
//	AILog.Info("Heap: insert: node"+node.f );
	list.append(node);
	local hole = list.len();
	for( ; hole > 1 && node.f < list[ (hole / 2)-1 ].f; hole /= 2 )
		list[ hole - 1 ] = list[ (hole / 2)-1 ];
	list[ hole - 1 ] = node;
//	 Dump();
}

/*------------------------------------------------------------------*/
function BinHeap::PopMin()
{	
	local min_node = null;
	if (list.len()) {
		min_node = list[ 0 ];
		list[0] = list[list.len()-1];
		BubbleDown();
		list.pop();
	}
//	Dump();
	return min_node;
}

/*------------------------------------------------------------------*/
function BinHeap::BubbleDown()
{
	local hole = 1;
   local tmp = list[ 0 ];
	
	while( hole * 2 <= list.len() ) {
		local child = hole * 2;
		if( (child != list.len() ) &&  list[ child ].f < list[ child - 1 ].f )
			child++;
		if( list[ child - 1 ].f < tmp.f )
			list[ hole - 1 ] = list[ child - 1 ];
		else
			break;
		hole = child;	
	}
	list[ hole - 1 ] = tmp;
}

/*------------------------------------------------------------------*/
function BinHeap::Exists(a)
{
	local i=0;
	foreach (list_node in list) {
		if (list_node.tile_index == a){
			return i;
		}
		i++;
	}
	return -1;
}

/*------------------------------------------------------------------*/
/** Just for debugging */
function BinHeap::Dump()
{
	local i=0;
	foreach (list_node in list) {
		AILog.Info("Heap: " + i + " " + list_node.f );
		i++;
	}
	return -1;
}


/*--------------------------------------------------------------------
|                                                                     |
|   PathFinder                                                        |
|                                                                     |
 --------------------------------------------------------------------*/
class PathFinder
{
	constructor() {
	}
	function BuildRoad(tile1, tile2);
}

/*------------------------------------------------------------------*/
function PathFinder::BuildRoad(tile1, tile2)
{
	if(tile1 == null || tile2 == null )	{
		AILog.Info("PathFinder: BuildByRoad: null tiles");
		return false;
	}

//	AISign.BuildSign(tile1, "tile1");
//	AISign.BuildSign(tile2, "tile2");

	local curr_tile = [tile1, tile2];

	if( !AIMap.IsValidTile(curr_tile[0]) || !AIMap.IsValidTile(curr_tile[1]))
		return false;

	local curr_node = Node();
	local open_list = BinHeap();
   local closed_list = NodeList();

	curr_node.tile_index = curr_tile[0];
	curr_node.g = 0; 
	curr_node.h =  10 * AITile.GetDistanceManhattanToTile(curr_tile[0],curr_tile[1]); // shortest distance to target 
	curr_node.f = curr_node.h; 
	curr_node.predec = 0;
	local k=0;
	while ( ( curr_tile[0] != curr_tile[1] ) && curr_node){	
		k++;
		if ( k % 100 == 0) 
			AIController.Sleep(1);

		local neighbour_tiles = Tile.GetAdjacentTiles(curr_tile[0]);
			
		neighbour_tiles.Valuate(AITile.IsBuildable);
		neighbour_tiles.KeepValue(1);

		for ( local tile = neighbour_tiles.Begin(); neighbour_tiles.HasNext(); tile = neighbour_tiles.Next()){
			local node = Node();
			node.tile_index = tile;
			node.g = curr_node.g + 10; // default cost
			node.h =  10 * AITile.GetDistanceManhattanToTile(tile, curr_tile[1]); // shortest distance to target 
			if (AIRoad.IsRoadTile(tile)){
				node.g -= 2;
			}
			if (AITile.GetSlope(tile)){
				node.g += 3;
			}
			node.predec = curr_node.tile_index;
			if (closed_list.Exists(tile) == -1){
				local cmd_ok = true;
				if (open_list.Exists(tile) == -1){
					if ( AITile.GetSlope(tile) || AITile.GetSlope(curr_tile[0])){
						if (curr_node.predec && ((curr_tile[0] - curr_node.predec) != (tile - curr_tile[0]))){
							cmd_ok = false;
						}
					} 
					if (cmd_ok) {
						open_list.Insert(node); 
					}
				}
			}
		}	
			
	   closed_list.Append(curr_node);
		curr_node = open_list.PopMin();
		if (curr_node) {
			curr_tile[0] = curr_node.tile_index;
		}
		else {
			return false;		
		}
	}
	
	local route_tiles = [];
	route_tiles.append(curr_node.tile_index);
//	AILog.Info("Add tile to route"+ AIMap.GetTileX(curr_node.tile_index)+" " + AIMap.GetTileY(curr_node.tile_index));
	local tile_node = Node(); 
	tile_node = curr_node;
	local finished = false;
	do {
	   local list_node = closed_list.Exists(tile_node.predec);
		if (list_node != -1){
			tile_node = list_node;
			route_tiles.append(tile_node.tile_index);
//			AILog.Info("Add tile to route"+ AIMap.GetTileX(tile_node.tile_index)+" " + AIMap.GetTileY(tile_node.tile_index));
			finished = (tile_node.predec == 0);
		}
		else{
			finished = true;
		}

	} while (!finished);

	local route_tile_prev = route_tiles.pop();
	if (route_tile_prev && route_tiles.len())	{
		do {
			local route_tile = route_tiles.pop();
			local built_road = AIRoad.BuildRoad(route_tile_prev, route_tile);
			if (!built_road) {
				if (AIError.GetLastError() != AIError.ERR_ALREADY_BUILT) {
					AILog.Info("Build road failed"+ AIError.GetLastErrorString()+" "+
						AIMap.GetTileX(route_tile_prev)+" " + AIMap.GetTileY(route_tile_prev)
				  		+" "+AIMap.GetTileX(route_tile)+" " + AIMap.GetTileY(route_tile)+" " + AITile.GetSlope(route_tile)+" " +
				  		AITile.GetSlope(route_tile_prev));
				}
			}
			route_tile_prev = route_tile;
		} while ( route_tiles.len());
	}	
	return true;
}


/*--------------------------------------------------------------------
|                                                                     |
|    TownManager                                                      |
|                                                                     |
 --------------------------------------------------------------------*/
class TownManager
{
	constructor() {

	}
	function CreateExitRoute(busstop, town1);
	function BuildBusStop(tile);
	function FindLineBusStopLocation(town, pass_cargo_id);
}

/*------------------------------------------------------------------*/
function TownManager::CreateExitRoute(busstop, town1) 
{
	local rtl = AITileList();
				
	rtl.Clear();
	rtl.AddRectangle(busstop + AIMap.GetTileIndex(-6,-6), busstop + AIMap.GetTileIndex(6,6));
	rtl.Valuate(AITile.IsBuildable);
	rtl.KeepValue(1);
	rtl.Valuate(AITile.GetDistanceManhattanToTile, AITown.GetLocation(town1));
	rtl.KeepBottom(1);
	local town_exit0 = rtl.Begin();
	local connected = PathFinder.BuildRoad(busstop, town_exit0); 
	if (!connected){
		AILog.Info("CreateExitRoute failed");
	}
	return town_exit0;

}

/*------------------------------------------------------------------*/
function TownManager::FindLineBusStopLocation(town, pass_cargo_id)
{
	local tl=AITileList();
	local aitile;
	local sl = null;
   local found = false;
	local tile = AITown.GetLocation(town);
	local x = AIMap.GetTileX(tile);
	local y = AIMap.GetTileY(tile);

	
	sl = AIStationList(AIStation.STATION_BUS_STOP);
	sl.Valuate(AIStation.GetDistanceManhattanToTile,tile);
	sl.KeepBottom(1);
	local station = sl.Begin();
	if (station){
		local dist = sl.GetValue(station);
		/* if we have a bus station in the town, use that one */
		if (dist < 15) {               
			return AIStation.GetLocation(station); 
		}
	}	


	tl.AddRectangle(AIMap.GetTileIndex(x-10,y-10), AIMap.GetTileIndex(x+10,y+10));
	if (tl.Count())
	{
		local tl2 = AITileList();
		tl2.AddRectangle(AIMap.GetTileIndex(x-10,y-10), AIMap.GetTileIndex(x+10,y+10));
		tl2.Valuate(AIRoad.IsRoadStationTile);
		tl2.KeepValue(1);
		for (local rstl = tl2.Begin();tl2.HasNext() ; rstl = tl2.Next()){
			 tl.RemoveRectangle(rstl+AIMap.GetTileIndex(-6,-6), rstl+AIMap.GetTileIndex(6,6));
		}	 


	   if (tl.Count()) {
	   	/* find all tiles that are next to a road tile */

	   	tl.Valuate(AIRoad.GetNeighbourRoadCount);	
   	   	tl.KeepAboveValue(0);
	   }
	   if (tl.Count()) {
	   	/* find all tiles that are not road */
	   	tl.Valuate(AIRoad.IsRoadTile);	
   	  	tl.KeepValue(0);
	   }
	   if (tl.Count()) {
	   	/* find all tiles that are not sloped */

	   	tl.Valuate(AITile.GetSlope);	
   	   tl.KeepValue(0);
	   }
   	if (tl.Count()) {
			tl.Valuate(AITile.GetCargoAcceptance, pass_cargo_id,
						 1, 1, 
						 AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP));
			aitile = tl.Begin();
			if (tl.GetValue(aitile) < 10){
				return null;	/* not enough acceptance, don't bother */			
			}
			
			for (;tl.HasNext() && !found; 
			    aitile = tl.Next()) {    
//				AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
				local adjacentTiles = Tile.GetAdjacentTiles(aitile);

				for(local tile2 = adjacentTiles.Begin(); 
			   	 adjacentTiles.HasNext() && !found; 
			    	 tile2 = adjacentTiles.Next()) {
			   	if(AIRoad.IsRoadTile(tile2) ) {
						if(!AIRoad.IsRoadStationTile(aitile)) {
			   			found = AITile.IsBuildable(aitile);
							if (!found)	{
								found = AITile.DemolishTile(aitile);
							}	
						}
					}
		   	}
				if (found) {
			   	break;
				}
			}
	   }
	}
	if (found)
		return aitile;	
	else
		return null;	
}

/*------------------------------------------------------------------*/
function TownManager::BuildBusStop(tile)
{
	local adjacentTiles = AITileList();
	local found = false;
	local success = false;

	local adjacentTiles = Tile.GetAdjacentTiles(tile);

	for(local tile2 = adjacentTiles.Begin(); 
			adjacentTiles.HasNext() && !found; 
			tile2 = adjacentTiles.Next()) {
		if(AIRoad.IsRoadTile(tile2) ) {
			AIRoad.BuildRoad(tile2, tile);
			local truck = false; 
			local drive_trough = false;
			success = AIRoad.BuildRoadStation(tile, tile2, truck, drive_trough);
			found = true;
		}
	}

	if (!success) {
		AILog.Info("Build busstop failed "+ AIError.GetLastErrorString());
	}
	return success;	
}

/*--------------------------------------------------------------------
|                                                                     |
|    RoutePlanner                                                     |
|                                                                     |
 --------------------------------------------------------------------*/
class RoutePlanner
{
	towns_used  = null;
	routes      = null;

	constructor() {
		this.towns_used = AIList();
		this.routes     = AIList();
	}
	function FindUnusedTowns();
}

/*------------------------------------------------------------------*/
function RoutePlanner::FindUnusedTowns()
{

	local town_list  = AITownList();
	local town_list2 = AITownList(); 

	local town    = 0;
	local town2   = 0;
	local town_it = AITown();
	local found   = false;
			
	town_list.RemoveList(this.towns_used);
	if (town_list.Count()){
		town_list.Valuate(AITown.GetPopulation);
//		town_list.Sort(AITownList.SORT_BY_VALUE, false);
		for (town_it = town_list.Begin(); town_list.HasNext(); 
	  		town_it = town_list.Next()){
			local tl2=AITileList();
			/* for our first town, be sure to use a town without a station */
			tl2.AddRectangle(AITown.GetLocation(town_it) + AIMap.GetTileIndex(-15,-15),
			   				AITown.GetLocation(town_it) + AIMap.GetTileIndex(15,15));
			tl2.Valuate(AIRoad.IsRoadStationTile);
			tl2.KeepValue(1);
			if (!tl2.Count()){
				town = town_it;
				this.towns_used.AddItem(town, 0);
				break;
			}
		}		

		town_list2.RemoveList(this.towns_used);
		town_list2.Valuate(AITown.GetDistanceManhattanToTile, AITown.GetLocation(town));
		town_list2.KeepBottom(6);
		town_list2.Valuate(AITown.GetPopulation);
		for (town_it = town_list2.Begin(); town_list2.HasNext(); 
   			town_it = town_list2.Next()){
			local tl2=AITileList();
			tl2.AddRectangle(AITown.GetLocation(town_it) + AIMap.GetTileIndex(-15,-15),
				   				AITown.GetLocation(town_it) + AIMap.GetTileIndex(15,15));
			tl2.Valuate(AIRoad.IsRoadStationTile);
			tl2.KeepValue(1);
			if (!tl2.Count()){
				town2 = town_it;
				break;
			}
		}		
		this.towns_used.AddItem(town2, 0);

		if (town2 > town){
			this.routes.AddItem((town2 << 10) + town, 0);
		}	
		else{
			this.routes.AddItem((town << 10) + town2, 0);
		}
	}
	else {
		return null;
	}
	return [ town, town2 ];
}

/*--------------------------------------------------------------------
|                                                                     |
|    Line                                                             |
|                                                                     |
 --------------------------------------------------------------------*/

class Line
{
	towns    = null;
	station  = null;
	depot    = null;
	vehicles = null;
	date_last_vehicle = 0;
	failed = false;
	passenger_cargo_id = 0;
	
	constructor(cargo_id) {
		towns     = [];
		station  = [null, null];
		depot    = null;
		vehicles  = [];
		failed = false;
		date_last_vehicle = 0;
		passenger_cargo_id = cargo_id;
	}
	function AddVehicles();
	function EstimateBusesNeeded();
	function AddDepot();
	function CreateNewLine(routeplanner);
}

function Line::CreateNewLine(routeplanner)
{
	local connected = true;
	local town_exit = [0,0];
	local success = [false,false];

	towns = routeplanner.FindUnusedTowns(); 

	if(towns && towns.len() == 2)
	{
		AILog.Info("New line from: " + AITown.GetName(towns[0]) + " to " + AITown.GetName(towns[1]));

		local town_idx = 0;
		foreach (town in towns){
			station[town_idx] =	TownManager.FindLineBusStopLocation(towns[town_idx],
																				passenger_cargo_id);				
			if (station[town_idx]) {
				success[town_idx] = TownManager.BuildBusStop(station[town_idx]);
				if (success[town_idx]) {
					AILog.Info("busstop " + town_idx + " built");
					town_exit[town_idx] =  TownManager.CreateExitRoute(AIRoad.GetRoadStationFrontTile(station[town_idx]), towns[(town_idx+1) %2 ]);
					if (!town_exit[town_idx]) {
						success[town_idx] = false;
						AILog.Info("Create exit " + town_idx + " failed");
						break;
					}					
				}
			}
			else {
				break;
			}	
			town_idx++;
		}
					
		// Build depot

		if((success[0] == false) || (success[1] == false)) {
			AILog.Info("Failed to build bus stations");
			connected = false;
		}
		if (connected) {
			depot = AddDepot(station[0]);
		}
		if (depot) {
			AILog.Info("depot built");
		}
		else {
			AILog.Info("Failed to build stations");
			connected = false;
		}
		if (connected) {
			connected = PathFinder.BuildRoad(town_exit[0], town_exit[1]); 
		}
		
		// make sure the depot and station are connected
		if(connected) {
			connected = PathFinder.BuildRoad(AIRoad.GetRoadStationFrontTile(station[0]), AIRoad.GetRoadDepotFrontTile(depot)) ;
		} 
		else {
			AILog.Info("Connect exits failed "+ AIError.GetLastErrorString());
		}
		// make sure the depot and town_exit are connected for a nicer route
		if(connected){
			connected = PathFinder.BuildRoad(town_exit[0], AIRoad.GetRoadDepotFrontTile(depot)) ;
		} 
		else {
			AILog.Info("Connect busstop 0 and depot failed"+ AIError.GetLastErrorString());
		}	
	}	
	else {
		AILog.Info("no 2 towns to connect");
	}
	return connected;
}

/*------------------------------------------------------------------*/
function Line::AddVehicles()
{
	if(this.station.len() != 2)
	{
		return null;
	}
	AILog.Info("AddVehicles");
	local bus_model = null;
	local min_reliability = 85;
	local new_bus;
	local i;
	local n_buses = 2;

	local engine_list = AIEngineList(AIVehicle.VEHICLE_ROAD);

	local balance = AICompany.GetBankBalance(AICompany.MY_COMPANY);
	engine_list.Valuate(AIEngine.GetPrice);
	engine_list.KeepBelowValue(balance);

	engine_list.Valuate(AIEngine.GetCargoType)
	engine_list.KeepValue(passenger_cargo_id); 

	engine_list.Valuate(AIEngine.GetCapacity)
	engine_list.KeepTop(1);

	bus_model = engine_list.Begin();
	if(bus_model == null)
	{
		AILog.Info("BuyBuses: failed to find a bus to build");
		return false;
	}

	n_buses = EstimateBusesNeeded();	
	// Buy buses
	for(i = 0; i < n_buses; ++i)
	{
		new_bus = AIVehicle.BuildVehicle(this.depot, bus_model);

		if(i>0)
		{
			AIOrder.ShareOrders(new_bus, this.vehicles[0]);
		}
		this.vehicles.append(new_bus);
		this.date_last_vehicle = AIDate.GetCurrentDate();
	}

	
	if(this.vehicles[0] == null || !AIVehicle.IsValidVehicle(this.vehicles[0]))
		AILog.Info("Vehicle[0] is not valid!");

	AIOrder.AppendOrder(this.vehicles[0], this.depot, AIOrder.AIOF_SERVICE_IF_NEEDED);
	AIOrder.AppendOrder(this.vehicles[0], this.station[0], AIOrder.AIOF_NONE); 
	AIOrder.AppendOrder(this.vehicles[0], this.station[1], AIOrder.AIOF_NONE);
	
	for(i = 0; i < n_buses; ++i)
	{
		AIVehicle.StartStopVehicle(this.vehicles[i]);
	}
}

/*------------------------------------------------------------------*/
function Line::EstimateBusesNeeded()
{

	if(towns.len() != 2)
		return 0;
	
	local acceptance = AITile.GetCargoAcceptance(AITown.GetLocation(towns[0]), 
																passenger_cargo_id, 1, 1, 
																AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP)) + 
							AITile.GetCargoAcceptance(AITown.GetLocation(towns[1]), 
																passenger_cargo_id, 1, 1, 
							AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP)) ;
	local distance = AIMap.DistanceManhattan( AITown.GetLocation(towns[0]), AITown.GetLocation(towns[1]) );

	AILog.Info("EstimateBusesNeeded: distance:    " + distance + 
	 										 " acceptance:  " + acceptance);

	local num_bus = 1 + (acceptance / 20) + (distance / 20);

	AILog.Info("Buy " + num_bus + " buses");
	return num_bus;
}


/*------------------------------------------------------------------*/
function Line::AddDepot(tile)
{
	local tl     = AITileList();
	local x      = AIMap.GetTileX(tile);
	local y      = AIMap.GetTileY(tile);
	local aitile = 0;
	local found  = false;
	
	tl.AddRectangle(AIMap.GetTileIndex(x-10,y-10), AIMap.GetTileIndex(x+10,y+10));
	if (tl.Count())
	{
//	   tl.Valuate(AITile.IsBuildable);
//	   tl.KeepValue(1);
	   if (tl.Count()) {
	   	/* find all tiles that are next to a road tile */

	   	tl.Valuate(AIRoad.GetNeighbourRoadCount);	
   	   	tl.KeepAboveValue(0);
	   }
	   if (tl.Count()) {
	   	/* find all tiles that are not road */
	   	tl.Valuate(AIRoad.IsRoadTile);	
   	  	tl.KeepValue(0);
	   }
	   if (tl.Count()) {
	   	/* find all tiles that are not sloped */

	   	tl.Valuate(AITile.GetSlope);	
   	   tl.KeepValue(0);
	   }
   	if (tl.Count()) {
			// let's find a tile with low passenger acceptance for the depot
			tl.Valuate(AITile.GetCargoAcceptance, this.passenger_cargo_id, 
					1, 1, AIStation.GetCoverageRadius (AIStation.STATION_BUS_STOP));
			tl.KeepBottom(8);
		
			for (aitile = tl.Begin();tl.HasNext() && !found; 
			    aitile = tl.Next()) {    
//				AISign.BuildSign(aitile,"ex"+tl.GetValue(aitile));
				local adjacentTiles = Tile.GetAdjacentTiles(aitile);

				/*Loop through all adjacent tiles*/

				for(local tile2 = adjacentTiles.Begin(); 
			   	 adjacentTiles.HasNext() && !found; 
			    	 tile2 = adjacentTiles.Next()) {
			   	if(AIRoad.IsRoadTile(tile2) ) {
						if(!AIRoad.IsRoadStationTile(aitile)) {
			   			found = AITile.IsBuildable(aitile);
							if (!found)	{
								found = AITile.DemolishTile(aitile);
							}	
							if (found) {
     							AIRoad.BuildRoad(tile2, aitile);
								AIRoad.BuildRoadDepot(aitile, tile2)
							}
						}
					}
		   	}
				if (found) {
			   	break;
				}
			}
	   }
	}
	if (found) {
		return aitile;	
	}
	else {
		return null;
	}
}

/*--------------------------------------------------------------------
|                                                                     |
|    ConvoyAI                                                         |
|                                                                     |
 --------------------------------------------------------------------*/
class ConvoyAI extends AIController {
	company            = null;
	stop               = false;
	vehicle_to_depot   = {};
	route_planner      = null;
	lines              = [];
	name               = null;
	passenger_cargo_id = 0;

	constructor() {
		this.company  = AICompany();
		route_planner = RoutePlanner();
		name = null;
	}

	function Start();
	function Stop();

	function SetCompanyName();
	function ManageLoan();
	function HasMoney(money);
}

/*------------------------------------------------------------------*/
function ConvoyAI::HasMoney(money)
{
	AILog.Info("Has money: bal " + AICompany.GetBankBalance(AICompany.MY_COMPANY)) ;
	if (AICompany.GetBankBalance(AICompany.MY_COMPANY) + (AICompany.GetMaxLoanAmount() - this.company.GetLoanAmount()) > money) return true;
	return false;
}

/*------------------------------------------------------------------*/
function ConvoyAI::Start()
{
	this.Sleep(1);

	this.SetCompanyName();
	this.name = AICompany.GetCompanyName(AICompany.MY_COMPANY);

	local list = AICargoList();
	for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
		if (AICargo.HasCargoClass(i, AICargo.CC_PASSENGERS)) {
			this.passenger_cargo_id = i;
			break;
		}
	}
	local i = 0;
	while(!this.stop)
	{
		this.Sleep(1);
		if((i%200)==0)
		{
			AILog.Info("Manage loan and Vehicles");
			ManageLoan();
			ManageVehicles();
			// ... check if we can afford to build some stuff
			if(this.HasMoney(20000))
			{
				AILog.Info("Start building");		

				local line = Line(this.passenger_cargo_id);
				this.company.SetLoanAmount(AICompany.GetMaxLoanAmount());
				local connected = line.CreateNewLine(this.route_planner);
				if(connected) {	
					AILog.Info("New line completed");
					line.AddVehicles(); 
				}
				else {
					AILog.Info("Failed to add new line");
					line.failed = true;	
				}
				lines.append(line);
				// Pay back unused money
				ManageLoan();
			}
		}
		i++;
	}
}

/*------------------------------------------------------------------*/
function ConvoyAI::Stop()
{
	AILog.Info("ConvoyAI::Stop()");
	this.stop = true;
}

/*------------------------------------------------------------------*/
function ConvoyAI::SetCompanyName()
{
	if (!AICompany.SetCompanyName("Convoy")) {
		local i = 2;
		while (!AICompany.SetCompanyName("Convoy #" + i)) {
			i++;
		}
	}
}

/*------------------------------------------------------------------*/
function ConvoyAI::ManageLoan()
{
	local balance = this.company.GetBankBalance(AICompany.MY_COMPANY);
	local loan = this.company.GetLoanAmount();
	local pay_back_balance =  2 * this.company.GetLoanInterval(); 
	local loan_interval = this.company.GetLoanInterval();
	local pay_back = 0;
	
	while(( balance - pay_back >= pay_back_balance) 
			&& (loan - pay_back > 0))	{
		pay_back += loan_interval;
	}
	if (pay_back) {
		if(!this.company.SetLoanAmount(loan - pay_back))	{
			AILog.Info(this.name + " Failed to pay back");
		}
		else {
			AILog.Info("Paid back: " + pay_back);
		}
	}	
	AILog.Info("Loan: " + this.company.GetLoanAmount());
}


/*------------------------------------------------------------------*/
function ConvoyAI::ManageVehicles()
{
	if (AIDate.GetMonth(AIDate.GetCurrentDate()) > 6) {
	
		local list = AIVehicleList();
		list.Valuate(AIVehicle.GetAge);
		list.KeepAboveValue(500);
		list.Valuate(AIVehicle.GetProfitThisYear);
   
		for (local i = list.Begin(); list.HasNext(); i = list.Next()) {
			local profit = list.GetValue(i);

			if (profit < 0) {
				/* Send the vehicle to depot if we didn't do so yet */
				if (!this.vehicle_to_depot.rawin(i) || this.vehicle_to_depot.rawget(i) != true) {
					AILog.Info(this.name + ": [INFO] Sending " + i + " to depot as	profit is: " + profit );
					AIVehicle.SendVehicleToDepot(i);
					this.vehicle_to_depot.rawset(i, true);
				}
			}
			/* Try to sell it over and over till it really is in the depot */
			if (this.vehicle_to_depot.rawin(i) && this.vehicle_to_depot.rawget(i) == true) {
				if (AIVehicle.SellVehicle(i)) {
					AILog.Info(this.name + ": [INFO] Selling " + i + " as it finally is in a depot.");
					/* Check if we are the last one serving those airports; else sell the airports */

					this.vehicle_to_depot.rawdelete(i);
				}
			}
		}
	}

	if (!this.HasMoney(12000)) return;

	foreach (line_it in lines) {
		if ( AIStation.GetCargoWaiting(AIStation.GetStationID(line_it.station[0]),
																			passenger_cargo_id) +
	   	AIStation.GetCargoWaiting(AIStation.GetStationID(line_it.station[1]),
																		passenger_cargo_id) > 250) {
			AILog.Info(this.name + ": [INFO] extra buses needed");	
			if (AIDate.GetCurrentDate() - line_it.date_last_vehicle > 50) {
				local new_veh = AIVehicle.CloneVehicle (line_it.depot, 
																	line_it.vehicles[0], true);			
				AIVehicle.StartStopVehicle(new_veh);
				line_it.vehicles.append(new_veh);				
				line_it.date_last_vehicle = AIDate.GetCurrentDate();
			}
		}
	}
}


